import argparse
import pathlib
import sys

from np_datajoint.classes import DataJointSession

def parse_args() -> argparse.Namespace:
    
    parser = argparse.ArgumentParser()
    parser.add_argument('--session', type=str, required=True)
    parser.add_argument('--paths', nargs='+', type=str, required=True)
    parser.add_argument('--probes', type=str, required=True)
    parser.add_argument('--no-sorting', dest='without_sorting', action='store_true')
    
    args = parser.parse_args(sys.argv[1:])
    args.paths = tuple(pathlib.Path(p) for p in args.paths)
    return args


if __name__ == "__main__":   
    
    args = parse_args()
    print(args)
    session = DataJointSession(args.session)

    session.upload(
        paths=args.paths,
        probes=args.probes,
        without_sorting=args.without_sorting
        )


from __future__ import annotations

import datetime
import functools
import hashlib
import itertools
import json
import logging
import os
import pathlib
import platform
import re
import shutil
import subprocess
import sys
import tempfile
import time
import traceback
import uuid
from collections import defaultdict
from collections.abc import Iterable
from typing import (Generator, List, Literal, MutableSequence, Optional,
                    Sequence, Set, Tuple)

import datajoint as dj
import djsciops.authentication as dj_auth
import djsciops.axon as dj_axon
import djsciops.settings as dj_settings
import fabric
import IPython
import ipywidgets as ipw
import np_logging
import numpy as np
import pandas as pd
import requests

from np_datajoint import classes, config

sorting_status_last_checked: dict[Optional[int], datetime.datetime] = dict()

class SessionDirNotFoundError(ValueError):
    pass

# general -----------------------------------------------------------------------------#
def checksum(path: pathlib.Path) -> str:
    hasher = hashlib.md5
    blocks_per_chunk = 128    
    multi_part_threshold_gb = 0.2
    if path.stat().st_size < multi_part_threshold_gb * 1024 ** 3:
        return hasher(path.read_bytes()).hexdigest()
    hash = hasher()
    with open(path, 'rb') as f:
        for chunk in iter(
            lambda: f.read(hash.block_size*blocks_per_chunk), b""
        ):
            hash.update(chunk)
    return hash.hexdigest()

def checksums_match(paths: Iterable[pathlib.Path]) -> bool:
    checksums = tuple(checksum(p) for p in paths)
    return all(c == checksums[0] for c in checksums)
    
def copy(src:pathlib.Path, dest:pathlib.Path):
    if not pathlib.Path(dest).parent.exists():
        pathlib.Path(dest).parent.mkdir(parents=True, exist_ok=True)
    attempts = 0
    if dest.exists() and dest.is_symlink():
        dest.unlink()
    while (
        True if not dest.exists() else not checksums_match((src, dest))
    ):
        if attempts == 2:
            logging.debug(f"Failed to copy {src} to {dest} with checksum-validation after {attempts=}")
            return
        shutil.copy2(src,dest)
        attempts += 1
    logging.debug(f"Copied {src} to {dest} with checksum-validation")

def symlink(src:pathlib.Path, dest:pathlib.Path):
    if 'win' in sys.platform:
        # Remote to remote symlink creation is disabled by default
        subprocess.run('fsutil behavior set SymlinkEvaluation R2R:1')
    if not pathlib.Path(dest).parent.exists():
        pathlib.Path(dest).parent.mkdir(parents=True, exist_ok=True)
    if dest.exists():
        if dest.is_symlink() and dest.resolve() == src.resolve():
            logging.debug(f"Symlink already exists to {src} from {dest}")
            return
        dest.unlink()
    dest.symlink_to(src)
    logging.debug(f"Created symlink to {src} from {dest}")
                
def get_session_folder(path: str | pathlib.Path) -> str | None:
    """Extract [8+digit session ID]_[6-digit mouse ID]_[8-digit date
    str] from a string or path"""
    session_reg_exp = R"[0-9]{8,}_[0-9]{6}_[0-9]{8}"

    session_folders = re.findall(session_reg_exp, str(path))
    if session_folders:
        if not all(s == session_folders[0] for s in session_folders):
            logging.debug(
                f"Mismatch between session folder strings - file may be in the wrong folder: {path}"
            )
        return session_folders[0]
    return None

def dir_size(path: pathlib.Path) -> int:
    """Return the size of a directory in bytes"""
    if not path.is_dir():
        raise ValueError(f"{path} is not a directory")
    dir_size = 0
    dir_size += sum(
        f.stat().st_size
        for f in pathlib.Path(path).rglob("*")
        if pathlib.Path(f).is_file()
    )
    return dir_size


# local ephys-related (pre-upload) ----------------------------------------------------- #
def is_new_ephys_folder(path: pathlib.Path) -> bool:
    "Look for hallmarks of a v0.6.x Open Ephys recording"
    return bool(
        tuple(path.rglob("Record Node*"))
        and tuple(path.rglob("structure.oebin"))
        and tuple(path.rglob("settings*.xml"))
        and tuple(path.rglob("continuous.dat"))
    )

def is_valid_ephys_folder(path: pathlib.Path) -> bool:
    "Check a single dir of raw data for size, v0.6.x+ Open Ephys for DataJoint."
    if not path.is_dir():
        return False
    if not is_new_ephys_folder(path):
        return False
    if not dir_size(path) > 275 * 1024**3:  # GB
        return False
    return True


def is_valid_pair_split_ephys_folders(paths: Sequence[pathlib.Path]) -> bool:
    "Check a pair of dirs of raw data for size, matching settings.xml, v0.6.x+ to confirm they're from the same session and meet expected criteria."
    if not paths:
        return False

    if any(not is_valid_ephys_folder(path) for path in paths):
        return False

    check_session_paths_match(paths)
    check_xml_files_match([tuple(path.rglob("settings*.xml"))[0] for path in paths])

    size_difference_threshold_gb = 2
    dir_sizes_gb = tuple(round(dir_size(path) / 1024**3) for path in paths)
    diffs = (abs(dir_sizes_gb[0] - size) for size in dir_sizes_gb)
    if not all(diff <= size_difference_threshold_gb for diff in diffs):
        print(
            f"raw data folders are not within {size_difference_threshold_gb} GB of each other"
        )
        return False

    return True


def get_raw_ephys_subfolders(path: pathlib.Path) -> List[pathlib.Path]:
    """
    Return a list of raw ephys recording folders, defined as the root that Open Ephys
    records to, e.g. `A:/1233245678_366122_20220618_probeABC`.

    Does not include the path supplied itself - only subfolders
    """

    subfolders = set()

    for f in pathlib.Path(path).rglob("*_probe*"):

        if not f.is_dir():
            continue

        if any(
            k in f.name
            for k in [
                "_sorted",
                "_extracted",
                "pretest",
                "_603810_",
                "_599657_",
                "_598796_",
            ]
        ):
            # skip pretest mice and sorted/extracted folders
            continue

        if not is_new_ephys_folder(f):
            # skip old/non-ephys folders
            continue

        if (size := dir_size(f)) and size < 1024**3 * 50:
            # skip folders that aren't above min size threshold (GB)
            continue

        subfolders.add(f)

    return sorted(list(subfolders), key=lambda s: str(s))


# - If we have probeABC and probeDEF raw data folders, each one has an oebin file:
#     we'll need to merge the oebin files and the data folders to create a single session
#     that can be processed in parallel
def get_single_oebin_path(path: pathlib.Path) -> pathlib.Path:
    """Get the path to a single structure.oebin file in a folder of raw ephys data.

    - There's one structure.oebin per Recording* folder
    - Raw data folders may contain multiple Recording* folders
    - Datajoint expects only one structure.oebing file per Session for sorting
    - If we have multiple Recording* folders, we assume that there's one
        good folder - the largest - plus some small dummy / accidental recordings
    """
    if not path.is_dir():
        raise ValueError(f"{path} is not a directory")

    oebin_paths = list(path.rglob("*structure.oebin"))

    if len(oebin_paths) > 1:
        oebin_parents = [f.parent for f in oebin_paths]
        dir_sizes = [dir_size(f) for f in oebin_parents]
        return oebin_paths[dir_sizes.index(max(dir_sizes))]

    elif len(oebin_paths) == 1:
        return oebin_paths[0]

    else:
        raise FileNotFoundError(f"No structure.oebin found in {path}")

def check_xml_files_match(paths: Sequence[pathlib.Path]):
    """Check that all xml files are identical, as they should be for
    recordings split across multiple locations e.g. A:/*_probeABC, B:/*_probeDEF"""
    if not all(s == ".xml" for s in [p.suffix for p in paths]):
        raise ValueError("Not all paths are XML files")
    if not all(p.is_file() for p in paths):
        raise FileNotFoundError("Not all paths are files, or they do not exist")
    if not checksums_match(paths):
        raise ValueError("XML files do not match")

def check_session_paths_match(paths: Sequence[pathlib.Path]):
    sessions = [get_session_folder(path) for path in paths]
    if any(not s for s in sessions):
        raise ValueError(
            "Paths supplied must be session folders: [8+digit lims session ID]_[6-digit mouse ID]_[6-digit datestr]"
        )
    if not all(s and s == sessions[0] for s in sessions):
        raise ValueError("Paths must all be for the same session")

def get_local_remote_oebin_paths(
    paths: pathlib.Path | Sequence[pathlib.Path],
) -> Tuple[Sequence[pathlib.Path], pathlib.Path]:
    """
    A `structure.oebin` file specfies the relative paths for each probe's files in a
    raw data dir. For split recs, a different oebin file lives in each dir (ABC/DEF).

    To process a new session on DataJoint, a single structure.oebin file needs to be uploaded,
    and its dir on the server specified in the SessionDirectory table / `session_dir` key.
    
    Input one or more paths to raw data folders that exist locally for a single
    session, and this func returns the paths to the `structure.oebin` file for each local folder,
    plus the expected relative path on the remote server for a single combined .oebin file.

    We need to upload the folder containing a `settings.xml` file - two
    folders above the structure.oebin file - but with only the subfolders returned from this function.
    """

    if isinstance(paths, pathlib.Path):
        paths = (paths,)

    if not any(is_new_ephys_folder(path) for path in paths):
        raise ValueError("No new ephys folder found in paths")
    check_session_paths_match(paths)

    local_session_paths: set[pathlib.Path] = set()
    for path in paths:

        ephys_subfolders = get_raw_ephys_subfolders(path)

        if ephys_subfolders and len(paths) == 1:
            # parent folder supplied: we want to upload its subfolders
            local_session_paths.update(e for e in ephys_subfolders)
            break  # we're done anyway, just making this clear

        if ephys_subfolders:
            logging.warning(
                f"Multiple subfolders of raw data found in {path} - expected a single folder."
            )
            local_session_paths.update(e for e in ephys_subfolders)
            continue

        if is_new_ephys_folder(path):
            # single folder supplied: we want to upload this folder
            local_session_paths.add(path)
            continue

    local_oebin_paths = sorted(
        list(set(get_single_oebin_path(p) for p in local_session_paths)),
        key=lambda s: str(s),
    )

    local_session_paths_for_upload = [p.parent.parent.parent for p in local_oebin_paths]
    # settings.xml file should be the same for _probeABC and _probeDEF dirs
    if len(local_session_paths_for_upload) > 1:
        check_xml_files_match([p / "settings.xml" for p in local_session_paths_for_upload])

    # and for the server we just want to point to the oebin file from two levels above -
    # shouldn't matter which oebin path we look at here, they should all have the same
    # relative structure
    remote_oebin_path = local_oebin_paths[0].relative_to(
        local_oebin_paths[0].parent.parent.parent
    )

    return local_oebin_paths, remote_oebin_path


def create_merged_oebin_file(
    paths: Sequence[pathlib.Path], probes: Sequence[str] = config.DEFAULT_PROBES
) -> pathlib.Path:
    """Take paths to two or more structure.oebin files and merge them into one.

    For recordings split across multiple locations e.g. A:/*_probeABC, B:/*_probeDEF
    """
    if isinstance(paths, pathlib.Path):
        return paths
    if any(not p.suffix == ".oebin" for p in paths):
        raise ValueError("Not all paths are .oebin files")
    if (
        len(paths) == 1
        and isinstance(paths[0], pathlib.Path)
        and paths[0].suffix == ".oebin"
    ):
        return paths[0]

    # ensure oebin files can be merged - if from the same exp they will have the same settings.xml file
    check_xml_files_match(
        [p / "settings.xml" for p in [o.parent.parent.parent for o in paths]]
    )

    logging.debug(f"Creating merged oebin file with {probes=} from {paths}")
    merged_oebin: dict = {}
    for oebin in sorted(paths):

        with open(oebin, "r") as f:
            oebin_data = json.load(f)

        for key in oebin_data:

            if merged_oebin.get(key, None) == oebin_data[key]:
                continue

            # 'continuous', 'events', 'spikes' are lists, which we want to concatenate across files
            if isinstance(oebin_data[key], List):
                for item in oebin_data[key]:
                    if merged_oebin.get(key, None) and item in merged_oebin[key]:
                        continue
                    # skip probes not specified in input args (ie. not inserted)
                    if "probe" in item.get(
                        "folder_name", ""
                    ):  # one is folder_name:'MessageCenter'
                        if not any(
                            f"probe{letter}" in item["folder_name"] for letter in probes
                        ):
                            continue
                    if merged_oebin.get(key, None) is None:
                        merged_oebin[key] = [item]
                    else:
                        merged_oebin[key].append(item)

    if not merged_oebin:
        raise ValueError("No data found in structure.oebin files")
    merged_oebin_path = pathlib.Path(tempfile.gettempdir()) / "structure.oebin"
    with open(str(merged_oebin_path), "w") as f:
        json.dump(merged_oebin, f, indent=4)

    return merged_oebin_path


# datajoint-related --------------------------------------------------------------------


def wait_on_process(sec=3600, msg="Still processing..."):
    fmt = "%a %H:%M"  # e.g. Mon 12:34
    file = sys.stdout
    time_now = time.strftime(fmt, time.localtime())
    time_next = time.strftime(fmt, time.localtime(time.time() + float(sec)))
    file.write("\n%s: %s\nNext check: %s\r" % (time_now, msg, time_next))
    file.flush()
    time.sleep(sec)


def add_new_ks_paramset(
    params: dict,
    description: str, 
    clustering_method: Literal['kilosort2','kilosort2.5','kilosort3'], 
):

    def dict_to_uuid(key):
        "Given a dictionary `key`, returns a hash string as UUID."
        hash = hashlib.md5()
        for k, v in sorted(key.items()):
            hash.update(str(k).encode())
            hash.update(str(v).encode())
        return uuid.UUID(hex=hash.hexdigest())

    param_dict = {
        "paramset_idx": max(config.AVAILABLE_PARAMSET_IDX)+1,
        "params": params,
        "paramset_desc": description,
        "clustering_method": clustering_method,
        "param_set_hash": dict_to_uuid(
            {**params, "clustering_method": clustering_method}
        ),
    }
    config.DJ_EPHYS.ClusteringParamSet.insert1(param_dict, skip_duplicates=True)


def get_clustering_parameters(
    paramset_idx: int = config.DEFAULT_KS_PARAMS_INDEX,
) -> Tuple[str, dict]:
    "Get description and dict of parameters from paramset_idx."
    return (config.DJ_EPHYS.ClusteringParamSet & {"paramset_idx": paramset_idx}).fetch1(
        "params"
    )


def all_sessions() -> dj.Table:
    "Correctly formatted sessions on Datajoint."
    logging.debug("Fetching all correctly-formatted sessions from DataJoint server")
    all_sessions = config.DJ_SESSION.Session.fetch()
    session_str_match_on_datajoint = lambda x: bool(
        get_session_folder(f"{x[1]}_{x[0]}_{x[2].strftime('%Y%m%d')}")
    )
    return (
        config.DJ_SESSION.Session
        & all_sessions[list(map(session_str_match_on_datajoint, all_sessions))]
    )


def get_sorting_status_all_sessions(
    paramset_idx: Optional[int] = config.DEFAULT_KS_PARAMS_INDEX,
) -> dj.Table:
    """Summary of processing for all probes across all sessions, with optional restriction on paramset_idx - modified from Thinh@DJ.
    - `paramset_idx = None` returns all probes
    - `paramset_idx = -1` returns probes with no paramset_idx (ie. haven't started
      processing)

    Table is returned, can be further restricted with queries.
    """

    def paramset_restricted(schema: dj.schemas.Schema) -> dj.schemas.Schema:
        "Restrict table to sessions that used one or more paramset_idx."
        if paramset_idx is None:
            return schema
        if -1 == paramset_idx:
            return schema & {"paramset_idx": None}
        return schema & {"paramset_idx": paramset_idx}

    logging.debug(
        f'Restricting processing status summary to sessions with paramset_idx={paramset_idx if paramset_idx is not None else "all"}'
    )

    @functools.cache
    def _get_sorting_status_all_sessions(paramset_idx) -> dj.Table:
        "Calling all these tables is slow (>10s), so cache the result."
        sorting_status_last_checked[paramset_idx] = datetime.datetime.now()

        session_process_status = all_sessions()

        session_process_status *= config.DJ_SESSION.Session.aggr(
            config.DJ_EPHYS.ProbeInsertion,
            probes="count(insertion_number)",
            keep_all_rows=True,
        )
        session_process_status *= config.DJ_SESSION.Session.aggr(
            config.DJ_EPHYS.EphysRecording,
            ephys="count(insertion_number)",
            keep_all_rows=True,
        )
        session_process_status *= config.DJ_SESSION.Session.aggr(
            config.DJ_EPHYS.LFP, lfp="count(insertion_number)", keep_all_rows=True
        )
        session_process_status *= config.DJ_SESSION.Session.aggr(
            paramset_restricted(config.DJ_EPHYS.ClusteringTask),
            task="count(insertion_number)",
            keep_all_rows=True,
        )
        session_process_status *= config.DJ_SESSION.Session.aggr(
            paramset_restricted(config.DJ_EPHYS.Clustering),
            clustering="count(insertion_number)",
            keep_all_rows=True,
        )
        session_process_status *= config.DJ_SESSION.Session.aggr(
            paramset_restricted(config.DJ_EPHYS.QualityMetrics),
            metrics="count(insertion_number)",
            keep_all_rows=True,
        )
        session_process_status *= config.DJ_SESSION.Session.aggr(
            paramset_restricted(config.DJ_EPHYS.WaveformSet),
            waveform="count(insertion_number)",
            keep_all_rows=True,
        )
        session_process_status *= config.DJ_SESSION.Session.aggr(
            paramset_restricted(config.DJ_EPHYS.WaveformSet),
            curated="count(insertion_number)",
            pidx_done='GROUP_CONCAT(insertion_number SEPARATOR ", ")',
            keep_all_rows=True,
        )
        return session_process_status.proj(
            ..., all_done="probes > 0 AND waveform = task"
        )

    if sorting_status_last_checked.get(
        paramset_idx, datetime.datetime.now()
    ) < datetime.datetime.now() - datetime.timedelta(hours=1):
        logging.debug("Clearing cache for _get_sorting_status_all_sessions")
        _get_sorting_status_all_sessions.cache_clear()
    return _get_sorting_status_all_sessions(paramset_idx)


def sorting_summary(
    paramset_idx: Optional[int] = config.DEFAULT_KS_PARAMS_INDEX,
) -> pd.DataFrame:
    """Summary of processing for all probes across all sessions, with optional restriction on `paramset_idx` - modified from Thinh@DJ.
    - `paramset_idx = None` returns all probes
    - `paramset_idx = -1` returns probes with no paramset_idx (ie. haven't started
      processing)
    """
    df = pd.DataFrame(get_sorting_status_all_sessions(paramset_idx))
    # make new 'session' column that matches our local session folder names
    session_str_from_datajoint_keys = (
        lambda x: x.session_id.astype(str)
        + "_"
        + x.subject
        + "_"
        + x.session_datetime.dt.strftime("%Y%m%d")
    )
    df = df.assign(session=session_str_from_datajoint_keys)
    # filter for sessions with correctly formatted session/mouse/date keys
    df = df.loc[~(pd.Series(map(get_session_folder, df.session)).isnull())]
    df.set_index("session", inplace=True)
    df.sort_values(by="session", ascending=False, inplace=True)
    # remove columns that were concatenated into the new 'session' column
    df.drop(columns=["session_id", "subject", "session_datetime", "lfp"], inplace=True)
    return df


def sorted_sessions(*args, **kwargs) -> Iterable[classes.DataJointSession]:
    df = sorting_summary(*args, **kwargs)
    yield from (
        classes.DataJointSession(session) for session in df.loc[df["all_done"] == 1].index
    )

def update_metrics_csv_with_missing_columns(path_or_session_folder: str | pathlib.Path, probe_idx: int, paramset_idx: int):
    probe = classes.ProbeDataJoint(session=path_or_session_folder, probe_letter=chr(probe_idx + ord('A')), paramset_idx=paramset_idx)
    if 'quality' in pd.read_csv(probe.metrics_csv).columns:
        logging.debug(f'{probe.metrics_csv} already contains columns added from DataJoint tables')
        return
    path = str(probe.metrics_csv)
    probe.metrics_df.to_csv(path)
    logging.debug(f'updated {probe.metrics_csv} with missing columns from DataJoint tables')

def copy_files_from_raw_to_sorted(path_or_session_folder: str | pathlib.Path, probe_idx: int, paramset_idx: int, make_symlinks=False, original_ap_continuous_dat=True):
    """Copy/rename/modify files to recreate extracted folders from Open Ephys
    pre-v0.6.
    
    Instead of making duplicate copies of files, symlinks can be made wherever possible.
    Lims upload copy utility won't follow symlinks, so the links should be made real
    before upload (possibly by running this again with symlinks disabled).
    
    If we skipped download of large files from DataJoint (default behavior), we have no AP
    `continuous.dat` file available for generating probe noise plots in QC.
    As an alternative, we can use the original, pre-median-subtraction file from the
    raw data folder (always symlinked due to size, and lims upload doesn't apply).  
    """
    self = classes.DataJointSession(path_or_session_folder)
    probe = chr(ord('A')+probe_idx)
    paths = self.get_raw_ephys_paths(probes = probe)
    local_oebin_paths, *_ = get_local_remote_oebin_paths(paths)
    assert len(local_oebin_paths) == 1
        
    raw: pathlib.Path = local_oebin_paths[0].parent
    sorted: pathlib.Path = self.dj_sorted_local_probe_path(probe_idx, paramset_idx)
    
    # Copy small files --------------------------------------------------------------------- #
    src_dest = []
    src_dest.append((
        raw / f"events/Neuropix-PXI-100.Probe{probe}-AP/TTL/states.npy",
        sorted / f"events/Neuropix-PXI-100.0/TTL_1/channel_states.npy",
    ))
    src_dest.append((
        raw / f"events/Neuropix-PXI-100.Probe{probe}-AP/TTL/sample_numbers.npy",
        sorted / f"events/Neuropix-PXI-100.0/TTL_1/event_timestamps.npy",
    ))
    src_dest.append((
        raw / f"events/Neuropix-PXI-100.Probe{probe}-AP/TTL/sample_numbers.npy",
        sorted / f"events/Neuropix-PXI-100.0/TTL_1/sample_numbers.npy",
    ))
    src_dest.append((
        raw / f"events/Neuropix-PXI-100.Probe{probe}-AP/TTL/full_words.npy",
        sorted / f"events/Neuropix-PXI-100.0/TTL_1/full_words.npy",
    ))
    src_dest.append((
        raw / f"continuous/Neuropix-PXI-100.Probe{probe}-AP/timestamps.npy",
        sorted / f"continuous/Neuropix-PXI-100.0/ap_timestamps.npy",
    ))
    src_dest.append((
        raw / f"continuous/Neuropix-PXI-100.Probe{probe}-LFP/continuous.dat",
        sorted / f"continuous/Neuropix-PXI-100.1/continuous.dat",
    ))
    src_dest.append((
        raw / f"continuous/Neuropix-PXI-100.Probe{probe}-LFP/timestamps.npy",
        sorted / f"continuous/Neuropix-PXI-100.1/lfp_timestamps.npy",
    ))
    
    logging.debug(f"{self.session_folder} probe{probe}: copying and renaming selected files from original raw data dir to downloaded sorted data dir")
    for src, dest in src_dest:
        symlink(src, dest) if make_symlinks else copy(src, dest)
                    
    # Fix Open Ephys v0.6.x event timestamps ----------------------------------------------- #
    # see https://gist.github.com/bjhardcastle/e972d59f482a549f312047221cd8eccb
    # check we haven't already applied the operation
    original = raw / f"events/Neuropix-PXI-100.Probe{probe}-AP/TTL/sample_numbers.npy"
    modified = sorted / f"events/Neuropix-PXI-100.0/TTL_1/event_timestamps.npy"
    if not modified.exists() or modified.is_symlink() or checksums_match((original, modified)):
        try:
            modified.unlink()
            modified.touch()
        except OSError:
            pass
        logging.debug(f"{self.session_folder} probe{probe}: adjusting `sample_numbers.npy` from OpenEphys and saving as `event_timestamps.npy`")
    
        src = raw / f"continuous/Neuropix-PXI-100.Probe{probe}-AP/sample_numbers.npy"
        continuous_sample_numbers = np.load(src, mmap_mode='r')
        first_sample = continuous_sample_numbers[0]
        
        event_timestamps = np.load(original.open('rb'))
        event_timestamps -= first_sample
        with modified.open('wb') as f:
            np.save(f, event_timestamps)
    
    # Create symlink to original AP data sans median-subtraction ----------------------------- #
    if original_ap_continuous_dat:
        src_dest = []
        src_dest.append((
            raw / f"continuous/Neuropix-PXI-100.Probe{probe}-AP/continuous.dat",
            sorted / f"continuous/Neuropix-PXI-100.0/continuous.dat",
        ))
        logging.debug(f"{self.session_folder} probe{probe}: copying original AP continuous.dat to downloaded sorted data dir")
        for src, dest in src_dest:
            symlink(src, dest)

def database_diagram() -> IPython.display.SVG:
    diagram = (
        dj.Diagram(config.DJ_SUBJECT.Subject)
        + dj.Diagram(config.DJ_SESSION.Session)
        + dj.Diagram(config.DJ_PROBE)
        + dj.Diagram(config.DJ_EPHYS)
    )
    return diagram.make_svg()


def is_hab(session_folder: pathlib.Path) -> Optional[bool]:
    "Return True/False, or None if not enough info to determine"
    for platform_json in session_folder.glob("*_platformD1.json"):
        if "habituation" in platform_json.read_text():
            return True
        return False
    return None

def session_upload_from_acq_widget() -> ipw.AppLayout:

    sessions = []
    if config.RUNNING_ON_ACQ:
        folders = get_raw_ephys_subfolders(pathlib.Path("A:")) + get_raw_ephys_subfolders(
            pathlib.Path("B:")
        )
        sessions = [
            get_session_folder(folder) 
            for folder in folders 
            if get_session_folder(folder) and is_new_ephys_folder(folder)
        ]
        for session in sessions:
            if sessions.count(session) < 2:
                sessions.remove(session) # folders that aren't split across A:/B:
    
    if not sessions:
        for f in filter(get_session_folder, config.NPEXP_PATH.iterdir()):
            if (
                f.name == get_session_folder(f.name) # excl dir names that have '- copy' appeneded
                and datetime.datetime.strptime(f.name.split('_')[-1], '%Y%m%d') > (datetime.datetime.now() - datetime.timedelta(days=14))
                and not is_hab(f)
            ):
                sessions.append(f.name)

    sessions = sorted(list(set(sessions)))
    probes_available_to_upload = 'ABCDEF'

    out = ipw.Output(layout={"border": "1px solid black"})

    session_dropdown = ipw.Dropdown(
        options=sessions,
        value=None,
        description="session",
        disabled=False,
    )

    upload_button = ipw.ToggleButton(
        description="Upload",
        disabled=True,
        button_style="",  # 'success', 'info', 'warning', 'danger' or ''
        tooltip="Upload raw data to DataJoint",
        icon="cloud-upload",  # (FontAwesome names without the `fa-` prefix)
    )

    progress_button = ipw.ToggleButton(
        description="Check sorting", 
        disabled=True,
        button_style="",  # 'success', 'info', 'warning', 'danger' or ''
        tooltip="Check sorting progress on DataJoint",
        icon="hourglass-half",  # (FontAwesome names without the `fa-` prefix)
    )

    surface_image = ipw.Image()
    surface_image.layout.height = '300px'
    surface_image.layout.visibility = 'hidden'
    
    def update_surface_image():
        for img in (config.NPEXP_PATH / session_dropdown.value).glob('*surface-image*'):
            if any(inserted_img in img.name for inserted_img in ('image4', 'image5')):
                break
        else:
            surface_image.layout.visibility = 'hidden'
            return
        with img.open('rb') as f:
            surface_image.value = f.read()
        surface_image.layout.visibility = 'visible'
            
    def handle_dropdown_change(change):
        if get_session_folder(change.new) is not None:
            upload_button.disabled = False
            upload_button.button_style = "warning"
            progress_button.disabled = False
            progress_button.button_style = "info"
            update_surface_image()
    session_dropdown.observe(handle_dropdown_change, names="value")

    def handle_upload_change(change):
        upload_button.disabled = True
        upload_button.button_style = "warning"
        with out:
            logging.info(f"Uploading probes: {probes_from_grid()}")
        session = classes.DataJointSession(session_dropdown.value)
        session.upload(probes=probes_from_grid())

    upload_button.observe(handle_upload_change, names="value")

    def handle_progress_change(change):
        with out:
            logging.info("Fetching summary from DataJoint...")
        progress_button.button_style = ""
        progress_button.disabled = True
        session = classes.DataJointSession(session_dropdown.value)
        try:
            with out:
                IPython.display.display(session.sorting_summary())
        except dj.DataJointError:
            logging.info(
                f"No entry found in DataJoint for session {session_dropdown.value}"
            )

    progress_button.observe(handle_progress_change, names="value")

    buttons = ipw.HBox([upload_button, progress_button])

    probe_select_grid = ipw.GridspecLayout(6, 1, grid_gap="0px")
    for idx, probe_letter in enumerate(probes_available_to_upload):
        probe_select_grid[idx, 0] = ipw.Checkbox(
            value=True,
            description=f"probe{probe_letter}",
            disabled=False,
            indent=True,
        )
    probe_select_and_image = ipw.HBox([surface_image, probe_select_grid])
    def probes_from_grid() -> str:
        probe_letters = ""
        for idx in range(6):
            if probe_select_grid[idx, 0].value == True:
                probe_letters += chr(ord("A") + idx)
        return probe_letters

    app = ipw.TwoByTwoLayout(
        top_right=probe_select_and_image,
        bottom_right=out,
        bottom_left=buttons,
        top_left=session_dropdown,
        width="100%",
        justify_items="center",
        align_items="center",
    )
    return IPython.display.display(app)

def local_dj_probe_pairs() -> Iterable[dict[str, classes.Probe]]:
    for session in sorted_sessions():
        for probe in config.DEFAULT_PROBES:
            try:
                local = classes.ProbeLocal(session, probe)
                dj = classes.ProbeDataJoint(session, probe)
                yield dict(local=local, dj=dj)
            except FileNotFoundError:
                continue



from __future__ import annotations

import datetime
import functools
import hashlib
import itertools
import json
import logging
import os
import pathlib
import platform
import re
import shutil
import subprocess
import sys
import tempfile
import time
import traceback
import uuid
from collections import defaultdict
from collections.abc import Iterable
from typing import (Generator, List, Literal, MutableSequence, Optional,
                    Sequence, Set, Tuple)

import datajoint as dj
import djsciops.authentication as dj_auth
import djsciops.axon as dj_axon
import djsciops.settings as dj_settings
import fabric
import IPython
import ipywidgets as ipw
import matplotlib.pyplot as plt
import np_logging
import numpy as np
import pandas as pd
import requests
import seaborn as sns

from np_datajoint import config, utils

logger = np_logging.getLogger(__name__)


class SessionDirNotFoundError(ValueError):
    pass

class DataJointSession:
    """A class to handle data transfers between local rigs/network shares, and the DataJoint server."""

    def __init__(self, path_or_session_folder: str | pathlib.Path):
        session_folder = utils.get_session_folder(str(path_or_session_folder))
        if session_folder is None:
            raise SessionDirNotFoundError(
                f"Input does not contain a session directory (e.g. 123456789_366122_20220618): {path_or_session_folder}"
            )
        self.session_folder = session_folder
        "[8+digit session ID]_[6-digit mouse ID]_[8-digit date]"
        if any(slash in str(path_or_session_folder) for slash in "\\/"):
            self.path = pathlib.Path(path_or_session_folder)
        else:
            self.path = None
        self.session_id, self.mouse_id, *_ = self.session_folder.split("_")
        self.date = datetime.datetime.strptime(
            self.session_folder.split("_")[2], "%Y%m%d"
        )
        try:
            if self.session_folder != self.session_folder_from_dj:
                raise SessionDirNotFoundError(
                    f"Session folder `{self.session_folder}` does not match components on DataJoint: {self.session_folder_from_dj}"
                )
        except dj.DataJointError:
            pass  # we could add metadata to datajoint here, but better to do that when uploading a folder, so we can verify session_folder string matches an actual folder
        logger.debug("%s initialized %s", self.__class__.__name__, self.session_folder)

    @property
    def session(self):
        "Datajoint session query - can be combined with `fetch` or `fetch1`"
        if not (session := config.DJ_SESSION.Session & {"session_id": self.session_id}):
            raise dj.DataJointError(f"Session {self.session_id} not found in database.")
        return session

    @property
    def session_key(self) -> dict[str, str | int]:
        "{subject:session_id}"
        return self.session.fetch1("KEY")

    @property
    def session_subject(self) -> str:
        return self.session.fetch1("subject")

    @property
    def session_datetime(self) -> datetime.datetime:
        return self.session.fetch1("session_datetime")

    @property
    def session_folder_from_dj(self) -> str:
        "Remote session dir re-assembled from datajoint table components. Should match our local `session_folder`"
        return f"{self.session_id}_{self.session_subject}_{self.session_datetime.strftime('%Y%m%d')}"

    @property
    def probe_insertion(self):
        return config.DJ_EPHYS.ProbeInsertion & self.session_key

    @property
    def clustering_task(self):
        return config.DJ_EPHYS.ClusteringTask & self.session_key

    @property
    def curated_clustering(self):
        "Don't get subtables from this query - they won't be specific to the session_key"
        return config.DJ_EPHYS.CuratedClustering & self.session_key

    @property
    def metrics(self):
        "Don't get subtables from this query - they won't be specific to the session_key"
        return config.DJ_EPHYS.QualityMetrics & self.session_key
    
    @property
    def sorting_finished(self) -> bool:
        return (
            len(self.clustering_task)
            == len(self.metrics)
            >= len(self.probe_insertion)
            > 0
        )

    @property
    def sorting_started(self) -> bool:
        return len(self.probe_insertion) > 0

    @property
    def remote_session_dir_relative(self) -> str:
        "Relative session_dir on datajoint server with no database prefix."
        return (config.DJ_SESSION.SessionDirectory & self.session_key).fetch1("session_dir")

    @property
    def remote_session_dir_outbox(self) -> str:
        "Root for session sorted data on datajoint server."
        return f"{config.DJ_OUTBOX}{'/' if not str(config.DJ_OUTBOX).endswith('/') else '' }{self.session_folder}/"

    @property
    def remote_session_dir_inbox(self) -> str:
        "Root for session uploads on datajoint server."
        return f"{config.DJ_INBOX}{'/' if not str(config.DJ_INBOX).endswith('/') else '' }{self.session_folder}/"

    def remote_sorted_probe_dir(self, probe_idx: int, paramset_idx: int) -> str:
        return (f"{config.DJ_OUTBOX}{'/' if not str(config.DJ_OUTBOX).endswith('/') else '' }" + 
                f"{(self.clustering_task & {'insertion_number':probe_idx, 'paramset_idx':paramset_idx}).fetch1('clustering_output_dir')}")
    
    @property
    def acq_paths(self) -> tuple[pathlib.Path, ...]:
        paths = []
        for drive, probes in zip("AB", ["_probeABC", "_probeDEF"]):
            path = pathlib.Path(f"{drive}:/{self.session_folder}{probes}")
            if path.is_dir():
                paths.append(path)
        return tuple(paths)

    @functools.cached_property
    def lims_info(self) -> Optional[dict]:
        response = requests.get(
            f"http://lims2/ecephys_sessions/{self.session_id}.json?"
        )
        if response.status_code != 200:
            return None
        return response.json()

    @property
    def lims_path(self) -> Optional[pathlib.Path]:
        if self.lims_info and (path := self.lims_info["storage_directory"]):
            return pathlib.Path('/'+path)
        return None

    @property
    def npexp_path(self) -> Optional[pathlib.Path]:
        path = config.NPEXP_PATH / self.session_folder
        return path if path.is_dir() else None

    @property
    def local_download_path(self) -> pathlib.Path:
        return pathlib.Path(config.LOCAL_INBOX)
    
    def dj_sorted_local_probe_path(self, probe_idx: int, sorted_paramset_idx: int = config.DEFAULT_KS_PARAMS_INDEX):
        return self.local_download_path / f"ks_paramset_idx_{sorted_paramset_idx}" / self.session_folder / f"{self.session_folder}_probe{chr(ord('A') + probe_idx)}_sorted"

    def npexp_sorted_probe_paths(
        self, probe_letter: Optional[str] = None
    ) -> pathlib.Path | Sequence[pathlib.Path]:
        "Paths to probe data folders sorted locally, with KS pre-2.0, or a single folder for a specified probe."
        path = lambda probe: pathlib.Path(
            Rf"//allen/programs/mindscope/workgroups/np-exp/{self.session_folder}/{self.session_folder}_probe{probe}_sorted/continuous/Neuropix-PXI-100.0"
        )
        if probe_letter is None or probe_letter not in config.DEFAULT_PROBES:
            return tuple(path(probe) for probe in config.DEFAULT_PROBES)
        else:
            return path(probe_letter)

    def add_clustering_task(
        self,
        paramset_idx: int = config.DEFAULT_KS_PARAMS_INDEX,
        probe_letters: Sequence[str] = config.DEFAULT_PROBES,
    ):
        "For existing entries in config.DJ_EPHYS.EphysRecording, create a new ClusteringTask with the specified `paramset_idx`"
        if not self.probe_insertion:
            logger.info(
                f"Probe insertions have not been auto-populated for {self.session_folder} - cannot add additional clustering task yet."
            )
            return
            # TODO need an additional check on reqd metadata/oebin file

        for probe_letter in probe_letters:
            probe_idx = ord(probe_letter) - ord("A")

            if (
                not config.DJ_EPHYS.EphysRecording
                & self.session_key
                & {"insertion_number": probe_idx}
            ):
                if (
                    config.DJ_EPHYS.ClusteringTask
                    & self.session_key
                    & {"insertion_number": probe_idx}
                ):
                    msg = f"ClusteringTask entry already exists - processing should begin soon, then additional tasks can be added."
                elif self.probe_insertion & {"insertion_number": probe_idx}:
                    msg = f"ProbeInsertion entry already exists - ClusteringTask should be auto-populated soon."
                else:
                    msg = f"ProbeInsertion and ClusteringTask entries don't exist - either metadata/critical files are missing, or processing hasn't started yet."
                logger.info(
                    f"Skipping ClusteringTask entry for {self.session_folder}_probe{probe_letter}: {msg}"
                )
                continue

            insertion_key = {
                "subject": self.mouse_id,
                "session": self.session_id,
                "insertion_number": probe_idx,
            }

            method = (
                (
                    config.DJ_EPHYS.ClusteringParamSet * config.DJ_EPHYS.ClusteringMethod
                    & insertion_key
                )
                .fetch("clustering_method")[paramset_idx]
                .replace(".", "-")
            )

            output_dir = f"{self.remote_session_dir_relative}/{method}_{paramset_idx}/probe{probe_letter}_sorted"

            task_key = {
                "subject": self.mouse_id,
                "session_id": self.session_id,
                "insertion_number": probe_idx,
                "paramset_idx": paramset_idx,
                "clustering_output_dir": output_dir,
                "task_mode": "trigger",
            }

            if config.DJ_EPHYS.ClusteringTask & task_key:
                logger.info(f"Clustering task already exists: {task_key}")
                return
            else:
                config.DJ_EPHYS.ClusteringTask.insert1(task_key, replace=True)

    def get_raw_ephys_paths(
        self,
        paths: Optional[Sequence[str | pathlib.Path]] = None,
        probes: str = config.DEFAULT_PROBES,
    ) -> tuple[pathlib.Path, ...]:
        """Return paths to the session's ephys data.
        The first match is returned from:
        1) paths specified in input arg,
        2) self.path
        3) A:/B: drives if running from an Acq computer
        4) session folder on lims
        5) session folder on npexp (should be careful with older sessions where data
            may have been deleted)
        """
        for path in (paths, self.path, self.acq_paths, self.lims_path, self.npexp_path):
            if not path:
                continue
            if not isinstance(path, Sequence):
                path = (path,)

            def filter(path):
                if match := re.search("(?<=_probe)[A-F]{,}", path.name):
                    if any(p in probes for p in match[0]):
                        return path

            matching_session_folders = {
                s for p in path for s in utils.get_raw_ephys_subfolders(p) if filter(s)
            }
            if len(matching_session_folders) == 1 or utils.is_valid_pair_split_ephys_folders(
                matching_session_folders
            ):
                logger.debug(matching_session_folders)
                return tuple(matching_session_folders)
        else:
            raise FileNotFoundError(f"No valid ephys raw data folders (v0.6+) found")

    def upload_from_hpc(self, paths, probes, without_sorting):
        "Relay job to HPC. See .upload() for input arg details."
        hpc_user = "svc_neuropix"
        logger.info(
            "Connecting to HPC via SSH, logs will continue in /allen/ai/homedirs/%s", hpc_user
        )
        with fabric.Connection(f"{hpc_user}@hpc-login") as ssh:

            # copy up-to-date files to hpc
            src_dir = pathlib.Path(__file__).parent / 'hpc'
            slurm_launcher = "slurm_launcher.py"
            upload_script = "upload_session.py"
            hpc_dir = "np_datajoint/"
            for file in (src_dir / slurm_launcher, src_dir / upload_script):
                ssh.put(file, hpc_dir)

            # launch upload job via slurm
            slurm_env = datajoint_env = "dj"
            cmd = f"conda activate {slurm_env};"
            cmd += f"cd {hpc_dir}; python {slurm_launcher}"# --env {datajoint_env}"
            # addtl args below will be passed to slurm batch job and parsed by upload_script
            cmd += f" {upload_script} --session {self.session_folder} --paths {' '.join(path.as_posix() for path in paths)} --probes {probes}"
            if without_sorting:
                cmd += " --without_sorting"
            ssh.run(cmd)  # submit all cmds in one line to enforce sequence -
            # ensures slurm job doesn't run until conda env is activated
            logger.debug(cmd)

    def upload(
        self,
        paths: Optional[Sequence[str | pathlib.Path]] = None,
        probes: Sequence[str] = config.DEFAULT_PROBES,
        without_sorting=False,
    ):
        """Upload from rig/network share to DataJoint server.

        Accepts a list of paths to upload or, if None, will try to upload from self.path,
        then A:/B:, then lims, then npexp.
        """
        paths = self.get_raw_ephys_paths(paths, probes)
        logger.debug(f"Paths to upload: {paths}")

        local_oebin_paths, remote_oebin_path = utils.get_local_remote_oebin_paths(paths)
        local_session_paths_for_upload = (
            p.parent.parent.parent for p in local_oebin_paths
        )

        data_on_network = str(paths[0]).startswith("//allen") or str(paths[0]).startswith("/allen")
        if data_on_network and not config.RUNNING_ON_HPC:
            try:
                self.upload_from_hpc(paths, probes, without_sorting)
                return
            except Exception as e:
                logger.exception(
                    "Could not connect to HPC: uploading data on //allen from local machine, which may be unstable"
                )

        if data_on_network or config.RUNNING_ON_HPC:
            config.BOTO3_CONFIG[
                "max_concurrency"
            ] = 100  # transfers over network can crash if set too high
            logger.info(
                "Data on network: limiting max_concurrency to %s",
                config.BOTO3_CONFIG["max_concurrency"],
            )

        if not without_sorting:
            self.create_session_entry(remote_oebin_path)

            # upload merged oebin file
            # ------------------------------------------------------- #
            temp_merged_oebin_path = utils.create_merged_oebin_file(local_oebin_paths, probes)
            dj_axon.upload_files(
                source=temp_merged_oebin_path,
                destination=f"{self.remote_session_dir_inbox}{remote_oebin_path.parent.as_posix()}/",
                session=config.S3_SESSION,
                s3_bucket=config.S3_BUCKET,
            )

        # upload rest of raw data
        # ------------------------------------------------------- #
        logger.getLogger("web").info(
            f"Started uploading raw data {self.session_folder}"
        )
        ignore_regex = ".*\.oebin"
        ignore_regex += "|.*\.".join(
            [" "]
            + [f"probe{letter}-.*" for letter in set(config.DEFAULT_PROBES) - set(probes)]
        ).strip()

        for local_path in local_session_paths_for_upload:
            dj_axon.upload_files(
                source=local_path,
                destination=self.remote_session_dir_inbox,
                session=config.S3_SESSION,
                s3_bucket=config.S3_BUCKET,
                boto3_config=config.BOTO3_CONFIG,
                ignore_regex=ignore_regex,
            )
        logger.getLogger("web").info(
            f"Finished uploading raw data {self.session_folder}"
        )
    
    @functools.cached_property
    def map_probe_idx_to_available_sorted_paramset_idx(self) -> dict[int, tuple[int]]:
        "May return empty tuple if no paramsets processed for probe."
        processed: list[list[int]] = self.curated_clustering.fetch('insertion_number', 'paramset_idx')
        probe_paramset_idx_mapping = defaultdict(tuple)
        for insertion_number in processed[0]:
            probe_paramset_idx_mapping[insertion_number] = tuple(paramset_idx for n, paramset_idx in enumerate(processed[1]) if processed[0][n] == insertion_number)
        return probe_paramset_idx_mapping
    
    def filtered_remote_and_local_sorted_paths(self, probe_idx: int, paramset_idx: int, skip_large_files=True) -> tuple[Sequence[str], Sequence[str]]:
        "Return list of sorted probe files on server and list of corresponding local paths, with mods to match internal sorting pipeline names/folder structure."
        large_file_threshold_gb = 1
        
        remote_sorted_probe_dir: pathlib.Path = self.remote_sorted_probe_dir(probe_idx, paramset_idx)
        local_sorted_probe_dir: pathlib.Path = self.dj_sorted_local_probe_path(probe_idx, paramset_idx)
        
        all_src_files: dict[str, str] = dj_axon.list_files(
            session=config.S3_SESSION,
            s3_bucket=config.S3_BUCKET,
            s3_prefix=remote_sorted_probe_dir,
            include_contents_hash=False,
            as_tree=False,
        )
        
        root = local_sorted_probe_dir
        ap = root / 'continuous/Neuropix-PXI-100.0'
        logs = root / 'logs'
        for folder in (root, ap, logs):
            folder.mkdir(exist_ok=True, parents=True)
            
        src = []
        dest = []
        def append(remote_file: pathlib.Path, remote_size:int, local_dir: pathlib.Path, local_name: str = None):
            dest_path = local_dir / (local_name or remote_file.name)
            if dest_path.exists() and dest_path.stat().st_size == remote_size:
                return
            src.append(remote_file.as_posix())
            dest.append(str(dest_path))
            
        src_paths = (pathlib.Path(src_file['key']) for src_file in all_src_files)
        src_sizes = (int(src_file['_size']) for src_file in all_src_files)
        for path, size in zip(src_paths, src_sizes):
            if skip_large_files and size > large_file_threshold_gb * 1024 ** 3:
                continue
            if '.' in path.stem or '.json.' in path.name:
                # log files with '.' in stem causing a FileNotFoundError on boto3 download
                continue
            if path.stem == 'probe_info':
                append(path, size, root)
            elif path.suffix in ('.json','.txt'):
                append(path, size, logs)
            elif path.stem == 'probe_depth':
                append(path, size, root, f"probe_depth_{chr(probe_idx + ord('A'))}.png")
            elif path.name == 'cluster_group.tsv':
                append(path, size, ap, 'cluster_group.tsv.v2')
            else:
                append(path, size, ap)
        return src, dest
            
    def download(self, paramset_idx: Optional[int] = config.DEFAULT_KS_PARAMS_INDEX, skip_large_files=True):
        "Download files from sorted probe dirs on server to local inbox."
        
        for probe_idx, sorted_paramset_idxs in self.map_probe_idx_to_available_sorted_paramset_idx.items():
            for sorted_paramset_idx in sorted_paramset_idxs:
                if paramset_idx and sorted_paramset_idx != paramset_idx:
                    continue
                logger.debug(f"Downloading sorted data for {self.session_folder} probe{chr(ord('A')+probe_idx)}, paramset_idx={sorted_paramset_idx}")
                src_list, dest_list = self.filtered_remote_and_local_sorted_paths(probe_idx, sorted_paramset_idx, skip_large_files)
                trailing_slash = '/' if 'win' not in sys.platform else '\\' # add if dest is dir
                for src, dest in zip(src_list, dest_list):    
                    config.S3_SESSION.s3.Bucket(config.S3_BUCKET).download_file(
                        Key=src,
                        Filename=dest,
                        Config=dj_axon.boto3.s3.transfer.TransferConfig(**config.BOTO3_CONFIG),
                    )
                try:    
                    utils.update_metrics_csv_with_missing_columns(self.session_folder, probe_idx, sorted_paramset_idx)
                except Exception as exc:
                    logger.exception(exc)
                utils.copy_files_from_raw_to_sorted(self.session_folder, probe_idx, sorted_paramset_idx, make_symlinks=True, original_ap_continuous_dat=True)
        logger.info(
            f"Finished downloading sorted data for {self.session_folder}"
        )

    def sorting_summary(self, *args, **kwargs) -> pd.DataFrame:
        df = utils.sorting_summary(*args, **kwargs)
        return df.loc[[self.session_folder]].transpose()

    def create_session_entry(self, remote_oebin_path: pathlib.Path):
        "Insert metadata for session in datajoint tables"

        remote_session_dir_relative = (
            pathlib.Path(self.session_folder) / remote_oebin_path.parent
        )

        if config.DJ_SESSION.SessionDirectory & {"session_dir": self.session_folder}:
            logger.info(f"Session entry already exists for {self.session_folder}")

        if not config.DJ_SUBJECT.Subject & {"subject": self.mouse_id}:
            # insert new subject
            config.DJ_SUBJECT.Subject.insert1(
                {
                    "subject": self.mouse_id,
                    "sex": "U",
                    "subject_birth_date": "1900-01-01",
                },
                skip_duplicates=True,
            )

        with config.DJ_SESSION.Session.connection.transaction:
            config.DJ_SESSION.Session.insert1(
                {
                    "subject": self.mouse_id,
                    "session_id": self.session_id,
                    "session_datetime": self.date,
                },
                skip_duplicates=True,
            )
            config.DJ_SESSION.SessionDirectory.insert1(
                {
                    "subject": self.mouse_id,
                    "session_id": self.session_id,
                    "session_dir": remote_session_dir_relative.as_posix() + "/",
                },
                replace=True,
            )


class Probe:
    
    def __init__(self, session:str|DataJointSession, probe_letter:str, **kwargs):
        if isinstance(session, DataJointSession):
            self.session = session
        else:
            self.session = DataJointSession(session)
        self.probe_letter = probe_letter
        
    @property
    def sorted_data_dir(self) -> pathlib.Path:
        "Local path to sorted data directory"
        raise NotImplementedError
    
    def __repr__(self):
        return f"{self.__class__.__name__}('{self.session.session_folder}', probe_letter='{self.probe_letter}')"
    
    @property
    def probe_idx(self) -> int:
        return ord(self.probe_letter) - ord('A')
    
    @property
    def metrics_csv(self) -> pathlib.Path:
        return self.sorted_data_dir / 'metrics.csv'
    
    @property
    def metrics_df(self) -> pd.DataFrame:
        if not self.metrics_csv.exists():
            raise FileNotFoundError(f"Does not exist: {self.metrics_csv}")
        if not hasattr(self, '_metrics_df'):
            self._metrics_df = pd.read_csv(self.metrics_csv)
            self._metrics_df.set_index('cluster_id', inplace=True) 
        return self._metrics_df
    
    def plot_metric_good_units(
        self, 
        metric:str, 
        ax:plt.Axes=None, 
        **kwargs
        ) -> plt.Axes | None:
        if not self.metrics_csv.exists():
            logger.info(f"No metrics.csv file found for {self:!r}")
            return None
        if 'quality' not in self.metrics_df.columns:
            logger.info(f"No quality column in metrics for {self:!r}")
            return None
        if all(self.metrics_df['quality'] == 'noise'):
            logger.info(f"All clusters are noise for {self:!r}")
            return None
        if len(self.metrics_df.loc[self.metrics_df['quality']=='good']) == 1:
            logger.info(f"Only one good cluster for {self:!r}")
            return None
        if ax is None:
            fig, ax = plt.subplots()
        sns.kdeplot(
            self.metrics_df[metric].loc[self.metrics_df['quality'] == 'good'],
            ax=ax,
            **kwargs)
        return ax

    @property
    def qc_units_df(self) -> pd.DataFrame:
        if not hasattr(self, '_qc_units_df'):
            self.get_qc_units()
        return self._qc_units_df
    
    def get_qc_units(self) -> None:
        "From `probeSync.load_spike_info`"
        spike_clusters = np.load(self.sorted_data_dir / 'spike_clusters.npy')
        spike_times = np.load(self.sorted_data_dir / 'spike_times.npy')
        templates = np.load(self.sorted_data_dir / 'templates.npy')
        spike_templates = np.load(self.sorted_data_dir / 'spike_templates.npy')
        channel_positions = np.load(self.sorted_data_dir / 'channel_positions.npy')
        amplitudes = np.load(self.sorted_data_dir / 'amplitudes.npy')
        unit_ids = np.unique(spike_clusters)

        # p_sampleRate = 30_000
        units = {}
        for u in unit_ids:
            ukey = str(u)
            units[ukey] = {}

            unit_idx = np.where(spike_clusters==u)[0]
            unit_sp_times = spike_times[unit_idx] #/p_sampleRate - shift
            
            units[ukey]['times'] = unit_sp_times
            
            #choose 1000 spikes with replacement, then average their templates together
            chosen_spikes = np.random.choice(unit_idx, 1000)
            chosen_templates = spike_templates[chosen_spikes].flatten()
            units[ukey]['template'] = np.mean(templates[chosen_templates], axis=0)
            units[ukey]['peakChan'] = np.unravel_index(np.argmin(units[ukey]['template']), units[ukey]['template'].shape)[1]
            units[ukey]['position'] = channel_positions[units[ukey]['peakChan']]
            units[ukey]['amplitudes'] = amplitudes[unit_idx]
            
        units_df = pd.DataFrame.from_dict(units, orient='index')
        units_df['cluster_id'] = units_df.index.astype(int)
        units_df = units_df.set_index('cluster_id')
        units_df = pd.merge(self.metrics_df, units_df, left_index=True, right_index=True, how='outer')
        units_df['probe'] = self.probe_letter
        units_df['uid'] = units_df['probe'] + units_df.index.astype(str)
        units_df = units_df.set_index('uid')
        units_df = units_df.loc[units_df['quality']=='good']
        self._qc_units_df = units_df
        
        
class ProbeLocal(Probe):
    def __init__(self,*args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.sorted_data_dir.exists():
            raise FileNotFoundError(f"Probe data expected at {self.sorted_data_dir}")
    
    @property
    def sorted_data_dir(self) -> pathlib.Path:
        return self.session.npexp_sorted_probe_paths(self.probe_letter)

    @property
    def depth_img(self) -> pathlib.Path:
        img = self.sorted_data_dir.parent.parent / f'probe_depth_{self.probe_letter}.png' 
        return img if img.exists() else self.sorted_data_dir.parent.parent / f'probe_depth.png'
    
    
class ProbeDataJoint(Probe):
    
    paramset_idx: int # 1 = KS2.0, 2 = KS2.5
    
    def __init__(self, *args, **kwargs):
        if (
            (paramset_idx := kwargs.pop('paramset_idx', config.DEFAULT_KS_PARAMS_INDEX)) is not None
            and str(paramset_idx).strip().isdigit()
        ):
            self.paramset_idx = int(paramset_idx)
        super().__init__(*args, **kwargs)
        if not self.sorted_data_dir.exists():
            raise FileNotFoundError(f"Probe data expected at {self.sorted_data_dir}")
            
    @property
    def sorted_data_dir(self) -> pathlib.Path:
        return self.session.dj_sorted_local_probe_path(self.probe_idx, self.paramset_idx) / 'continuous/Neuropix-PXI-100.0' 

    @property
    def depth_img(self) -> pathlib.Path:
        img = self.sorted_data_dir.parent.parent / f'probe_depth_{self.probe_letter}.png' 
        return img if img.exists() else self.sorted_data_dir.parent.parent / f'probe_depth.png'
    
    @property
    def metrics_table(self):
        query = {
            'insertion_number': self.probe_idx,
            'paramset_idx': self.paramset_idx,
        }
        query = query | self.session.session_key
        metrics = utils.config.DJ_EPHYS.QualityMetrics.Cluster & query
        # join column from other table with quality label
        metrics *= ( 
            utils.config.DJ_EPHYS.CuratedClustering.Unit & query
        ).proj('cluster_quality_label')
        return metrics
    
    @property
    def metrics_df(self) -> Optional[pd.DataFrame]:
        "CSV from DJ is missing some columns - must be fetched from DJ tables."
        if not self.metrics_csv.exists():
            raise FileNotFoundError(f"Does not exist: {self.metrics_csv}")
        if not hasattr(self, '_metrics_df'):
            # fetch quality column from DJ and rename
            quality = pd.DataFrame(
                self.metrics_table.proj(
                    cluster_id='unit',
                    quality='cluster_quality_label',
                )
            )
            metrics = pd.read_csv(self.metrics_csv)
            #  don't set_index('cluster_id'): we need col 'unnamed: 0' for qc back compat
            self._metrics_df = metrics.join(quality['quality'])
        return self._metrics_df
      

class DRPilot(DataJointSession):
    """A class to handle data transfers between local rigs/network shares, and the DataJoint server."""

    def __init__(self, session_folder_path: str | pathlib.Path):
        self.path = pathlib.Path(session_folder_path)
        self.session_folder = self.get_session_folder(self.path.name)
        "DRpilot_[6-digit mouse ID]_[8-digit date]"
        if self.session_folder is None:
            raise SessionDirNotFoundError(
                f"Input does not contain a session directory (e.g. DRpilot_366122_20220618): {session_folder_path}"
            )
            
        _, self.mouse_id, date = self.session_folder.split("_")
        self.date = datetime.datetime.strptime(date, "%Y%m%d").date()
        
        self.session_id = self.mouse_id + date
        
        try:
            if self.session_folder != self.session_folder_from_dj:
                raise SessionDirNotFoundError(
                    f"Session folder `{self.session_folder}` does not match components on DataJoint: {self.session_folder_from_dj}"
                )
        except dj.DataJointError:
            pass  # we could add metadata to datajoint here, but better to do that when uploading a folder, so we can verify session_folder string matches an actual folder

        logger.debug("%s initialized %s", self.__class__.__name__, self.session_folder)

    @staticmethod
    def get_session_folder(path: str | pathlib.Path) -> str | None:
        """Extract ["DRpilot_[6-digit mouse ID]_[8-digit date str] from a string or path.
        """
        session_reg_exp = R"DRpilot_[0-9]{6}_[0-9]{8}"

        session_folders = re.findall(session_reg_exp, str(path))
        return session_folders[0] if session_folders else None
    
    @property
    def session_folder_from_dj(self) -> str:
        "Remote session dir re-assembled from datajoint table components. Should match our local `session_folder`"
        return f"DRpilot_{self.session_subject}_{self.session_datetime.strftime('%Y%m%d')}"

    def lims_info(self) -> Optional[dict]:
        """Get LIMS info for this session. 
        """
        logger.warning("LIMS info not available: LIMS sessions weren't created for for DRPilot experiments.")



from __future__ import annotations

import datetime
import hashlib
import json
import logging
import os
import pathlib
import re
import sys
import tempfile
import time
import uuid
from collections.abc import Iterable
from typing import Generator, List, Optional, Sequence, Set, Tuple

import datajoint as dj
import djsciops.authentication as dj_auth
import djsciops.axon as dj_axon
import djsciops.settings as dj_settings
import IPython
import ipywidgets as ipw
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import PIL
import seaborn as sns

from np_datajoint import classes, config, utils

sns.set_style("whitegrid")
    
LOCAL_QC_ROOT = pathlib.Path("//allen/programs/braintv/workgroups/nc-ophys/corbettb/NP_behavior_pipeline/QC")
DJ_QC_ROOT = pathlib.Path(f"//allen/programs/mindscope/workgroups/dynamicrouting/datajoint/qc")

dj_qc_subdir = lambda paramset_idx: f"ks_paramset_idx_{paramset_idx}"
QC_LOCATIONS = {
    'Local KS 2.0': LOCAL_QC_ROOT,
    'DataJoint KS 2.0': DJ_QC_ROOT / dj_qc_subdir(1),
    'DataJoint KS 2.5': DJ_QC_ROOT / dj_qc_subdir(2),
}
 
def available_qc_sessions() -> tuple[tuple[str, ...], tuple[pathlib.Path, ...]]:
    sessions: set[str] = set()
    paths: set[pathlib.Path] = set()
    for paramset_idx in (1,2):
        sorted_data_repository = DJ_QC_ROOT.with_name('inbox') / dj_qc_subdir(paramset_idx)
        for path in sorted_data_repository.iterdir():
            if session := utils.get_session_folder(path):
                paths.add(path)
                sessions.add(session)
    return tuple(sorted(list(sessions))), tuple(sorted(list(paths)))

def available_qc_img_relpaths(session: pathlib.Path) -> Sequence[pathlib.Path]:
    "Probe-related images from QC run on Datajoint-sorted data"
    root = DJ_QC_ROOT / dj_qc_subdir(paramset_idx=1) / session
    names = ('probe', 'unit', 'receptive', )
    return tuple(p.relative_to(root) for p in root.rglob('*.png') if any(s in str(p).lower() for s in names)) 

def equivalent_qc_imgs(session: str, img_relpath: pathlib.Path) -> dict[str, pathlib.Path]:
    "Versions of the same QC image created from different sorting sessions (local KS 2.0, DataJoint KS 2.0, 2.5)"
    return {label : root / session / img_relpath for label, root in QC_LOCATIONS.items()}
    
def qc_img_compare() -> IPython.display.display:
    
    display_width_pix = 500
    
    sessions, paths = available_qc_sessions()
    
    imgs = tuple(ipw.Image() for _ in QC_LOCATIONS.keys())

    def update_img_displays(session:str, relpath: pathlib.Path):
        img_paths = equivalent_qc_imgs(session, relpath)
        for idx, path in enumerate(img_paths.values()):
            img = imgs[idx]
            if path.exists():
                with path.open('rb') as f:
                    img.value = f.read()
                    width, height = PIL.Image.open(f).size
                r = width/height
                if r < 1:
                    img.layout.height = f'{display_width_pix}px'
                    img.layout.width = f'{int(display_width_pix*r)}px'
                else:
                    img.layout.height = f'{int(display_width_pix/r)}px'
                    img.layout.width = f'{display_width_pix}px'
                img.layout.visibility = 'visible'
                img.path = path
            else:
                img.layout.visibility = 'hidden'
                img.path = None
                
    session_dropdown = ipw.Dropdown(
        options=sessions,
        value='1191631184_615047_20220714',
        description="session",
        disabled=False,
    )
    img_dropdown = ipw.Dropdown(
        value=pathlib.Path('receptive_fields/615047_20220714_C_RFs_by_depth.png'),
        options=available_qc_img_relpaths(session_dropdown.value),
        description="image",
        disabled=False,
    )

    def handle_session_change(change):
        if utils.get_session_folder(change.new):
            img_dropdown.options = available_qc_img_relpaths(session_dropdown.value)
    session_dropdown.observe(handle_session_change, names="value")

    def handle_img_change(change):
        if change.new is not None and img_dropdown.value:
            update_img_displays(session_dropdown.value, img_dropdown.value)
    img_dropdown.observe(handle_img_change, names="value")

    img_link = ipw.HTML(value='Open image: ')
    
    def update_img_link(path: Optional[pathlib.Path]):
        if not path:
            img_link.value = ''
            return
        img_link.value = f'<h>{(path)}</h>'
    
    tabs = ipw.Tab()
    tabs.children = imgs
    for idx, title in enumerate(QC_LOCATIONS.keys()):
        tabs.set_title(idx, title)
    tabs.layout=ipw.Layout(align_content='center',height=f'{display_width_pix}px',width=f'{display_width_pix}px')
    
    def handle_tab_change(change:None):
        update_img_link(imgs[tabs.selected_index].path)
    tabs.observe(handle_tab_change, type='change', names="selected_index")
    
    update_img_displays(session_dropdown.value, img_dropdown.value)
    update_img_link(imgs[tabs.selected_index].path)
    
    return IPython.display.display(
        ipw.VBox([session_dropdown, img_dropdown, tabs, img_link])   
    )
    


from __future__ import annotations

import datetime
import functools
import hashlib
import itertools
import json
import logging
import os
import pathlib
import platform
import re
import shutil
import subprocess
import sys
import tempfile
import time
import traceback
import uuid
from collections import defaultdict
from collections.abc import Iterable
from typing import (Final, Generator, List, Literal, MutableSequence, Optional,
                    Sequence, Set, Tuple)

import datajoint as dj
import djsciops.authentication as dj_auth
import djsciops.axon as dj_axon
import djsciops.settings as dj_settings
import fabric
import IPython
import ipywidgets as ipw
import matplotlib.pyplot as plt
import np_config
import np_logging
import numpy as np
import pandas as pd
import requests
import seaborn as sns

for log_name in ("Primary", "datajoint"):  # both datajoint packages
    logging.getLogger(log_name).disabled = True
    # logging.getLogger(log_name).propagate = False
    

# config ------------------------------------------------------------------------------
# get zookeeper config 
zk_config = np_config.from_zk("/projects/np_datajoint/defaults/configuration")

# configure datajoint session
dj.config.update(
    zk_config["datajoint"]
)  # dj.config is a custom class behaving as a dict - don't directly assign a dict

S3_SESSION = dj_auth.Session(
    aws_account_id=zk_config["djsciops"]["aws"]["account_id"],
    s3_role=zk_config["djsciops"]["s3"]["role"],
    auth_client_id=zk_config["djsciops"]["djauth"]["client_id"],
    # bearer_token=zk_config["djsciops"]["djauth"]["client_secret"],
)
S3_BUCKET: str = zk_config["djsciops"]["s3"]["bucket"]

DJ_INBOX: str = zk_config["sorting"][
    "remote_inbox"
]  # f"mindscope_dynamic-routing/inbox"
DJ_OUTBOX: str = zk_config["sorting"][
    "remote_outbox"
]  # f"mindscope_dynamic-routing/inbox"
LOCAL_INBOX = pathlib.Path(zk_config["sorting"]["local_inbox"])

BOTO3_CONFIG: dict = zk_config["djsciops"]["boto3"]

# create virtual datajoint modules for querying tables ---------------------------------- #
DJ_SUBJECT = dj.VirtualModule("subject", "mindscope_dynamic-routing_subject")
DJ_SESSION = dj.VirtualModule("session", "mindscope_dynamic-routing_session")
DJ_EPHYS = dj.VirtualModule("ephys", "mindscope_dynamic-routing_ephys")
DJ_PROBE = dj.VirtualModule("probe", "mindscope_dynamic-routing_probe")

DEFAULT_PROBES = "ABCDEF"

DEFAULT_KS_PARAMS_INDEX: int = zk_config["sorting"][
    "default_kilosort_parameter_set_index"
]  # 1=KS 2.0, 2=KS 2.5
AVAILABLE_PARAMSETS: np.ndarray = DJ_EPHYS.ClusteringParamSet.fetch()
AVAILABLE_PARAMSET_IDX = tuple(s[0] for s in AVAILABLE_PARAMSETS)

RUNNING_ON_HPC: bool = bool(
    re.match("n[0-9]{,}|hpc-login", HOSTNAME := platform.node())
)  # hpc nodes named 'n73'
logging.debug("%s Running on HPC: %s", HOSTNAME, RUNNING_ON_HPC)
RUNNING_ON_ACQ: bool = bool(
    re.match("np[0-9]{1}-acq", os.getenv("aibs_comp_id", '').lower())
)

NPEXP_PATH = pathlib.Path("//allen/programs/mindscope/workgroups/np-exp")


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['np_datajoint', 'np_datajoint.hpc']

package_data = \
{'': ['*']}

install_requires = \
['datajoint==0.13',
 'djsciops==1.4.1',
 'fabric>=2,<3',
 'ipywidgets>=7,<8',
 'np_logging',
 'np_session',
 'numpy>=1,<2',
 'pandas>=1,<2',
 'requests',
 'seaborn',
 'setuptools>=67.4.0,<68.0.0']

setup_kwargs = {
    'name': 'np-datajoint',
    'version': '0.0.5',
    'description': 'Tools for spike-sorting Mindscope neuropixels ecephys sessions on DataJoint, retrieving results and comparing with locally-sorted equivalents.',
    'long_description': 'Tools for spike-sorting Mindscope neuropixels ecephys sessions on DataJoint, retrieving results and comparing with locally-sorted equivalents.',
    'author': 'Ben Hardcastle',
    'author_email': 'ben.hardcastle@alleninstitute.org',
    'maintainer': 'Ben Hardcastle',
    'maintainer_email': 'ben.hardcastle@alleninstitute.org',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



"Simplified by running this script from the env required for the job - just needs `simple_slurm`."
import argparse
import logging
import os
import pathlib
import sys

import simple_slurm

try:
    import np_logging
except ImportError:
    pass # nice to have but not required

logger = logging.getLogger(__name__)

slurm = simple_slurm.Slurm(
    cpus_per_task=1,
    partition='braintv',
    job_name='dj_upload',
    time='14:00:00',
    # output=f'~/np_datajoint/logs/{simple_slurm.Slurm.JOB_ARRAY_MASTER_ID}_{simple_slurm.Slurm.JOB_ARRAY_ID}.out',
    mem_per_cpu='2gb',
)

# parser = argparse.ArgumentParser()
# parser.add_argument('env', help='path to the environment to run the slurm job in')
# args = vars(parser.parse_args())
# env = args.pop('env', None)
# if sys.argv[1] == '--env':
#     env = sys.argv[2]
#     args = sys.argv[3:]
# else:
#     args = sys.argv[1:]

slurm.sbatch(cmd := f"python {' '.join(sys.argv[1:])}")
logger.info('submitted: slurm.sbatch %s', cmd)
print(cmd)